import Task
from Timer import Timer
import threading
import os
import json
import uuid
from matrix_wrapper import send_message

# Tasks, Taskgroups, subgroups, each Task has Timer

class Task():
    ID_JSON_ID = "id"
    NAME_JSON_ID = "name"
    TIMERS_JSON_ID = "timers"
    SUBTASKS_JSON_ID = "subtasks"
    STATE_JSON_ID = "state"

    STATE_TICKED = "ticked"
    STATE_UNTICKED = "unticked"

    ERROR_CODE_OK = 0
    ERROR_CODE_ALREADY_PART_OF = 1

    _id: str = None
    _name: str = None
    _timers: list[Timer]
    _state: str = None
    _backup: bool = None
    _subtasks: list[Task] = None
    _shared_timer_callbacks: list = None
    _successful_timer_callbacks: list = None
    _cancel_timer_callbacks: list = None
    _removal_callbacks: list = None

    tasks_lock: threading.Lock = threading.Lock()
    tasks: list[Task] = []

    _tasks_backup_file_lock: threading.Lock = threading.Lock()
    _backup_path: str = os.environ["TASKS_PATH"]

    def __init__(self, name: str, id: str=None, backup: bool=True) -> None:
        self._name = name
        self._backup = backup
        self._state = Task.STATE_UNTICKED

        self._subtasks = []
        self._timers = []
        self._shared_timer_callbacks = []
        self._successful_timer_callbacks = []
        self._cancel_timer_callbacks = []
        self._removal_callbacks = []


        self._shared_timer_callbacks.append(self.timer_expired_cancelled)
        self._successful_timer_callbacks.append(self.default_timer_callback)        

        self._id = id if id is not None else str(uuid.uuid4())
        self._redump()

        with Task.tasks_lock:
            Task.tasks.append(self)

    def add_timer(self, duration: float, text: str=None, id: str=None):
        timer = Timer(duration, text, id=id, backup=False, cancel_callbacks=self.get_cancel_timer_callbacks(), successfull_callbacks=self.get_succ_timer_callbacks())
        timer.start()

        self._timers.append(timer)
        self._redump()

        return timer

    def link_timer(self, timer: Timer) -> int:
        if timer in self._timers:
            return Task.ERROR_CODE_ALREADY_PART_OF

        timer._backup = False
        timer._remove_backup()

        timer.add_callbacks(successful_callbacks=self.get_succ_timer_callbacks(), cancel_callbacks=self.get_cancel_timer_callbacks())

        self._timers.append(timer)

        self._redump()
        return Task.ERROR_CODE_OK

    def add_subtask(self, name: str, id: str=None):
        subtask = Task(name, id, self._backup)
        subtask.add_removal_callback(self.subtask_removed)
        
        self._subtasks.append(subtask)
        self._redump()

        return subtask

    def link_subtask(self, task: Task) -> int:
        if task in self.get_subtasks():
            return Task.ERROR_CODE_ALREADY_PART_OF
        task._backup = self._backup
        task.add_removal_callback(self.subtask_removed)
        task._redump()

        self._subtasks.append(task)
        self._redump()
        return Task.ERROR_CODE_OK

    def unlink_subtask(self, task: Task):
        if task in self._subtasks:
            task.remove_removal_callback(self.subtask_removed)
            self._subtasks.remove(task)
            self._redump()


    def tick(self):
        self._state = Task.STATE_TICKED
        for timer in self._timers:
            timer.cancel()
        
        for subtask in self.get_subtasks():
            subtask.tick()
            subtask._redump()

        self._redump()

    def untick(self):
        self._state = Task.STATE_UNTICKED
        self._redump()

    def get_subtasks(self):
        return self._subtasks

    def get_timers(self):
        return self._timers

    def get_state(self):
        return self._state

    def get_subtasks_deep(self):
        subtasks = []
        for subtask in self.get_subtasks():
            subtasks.append(subtask.get_subtasks_deep())

        return {self: subtasks}

    def get_subtask(self, id: str):
        for subtask in self.get_subtasks():
            if subtask.get_id() == id:
                return subtask

    def _checked_dump(self):
        if self._backup:
            self._dump()

    def _dump(self):
        dump_str = Task.get_dump_str(self.get_dump())
        with Task._tasks_backup_file_lock:
            with open(Task._backup_path, "a+") as f:
                f.write(dump_str)

    def get_dump(self):
        return {Task.ID_JSON_ID: str(self.get_id()),
        Task.NAME_JSON_ID: self.get_name(),
        Task.STATE_JSON_ID: self.get_state(),
        Task.TIMERS_JSON_ID: [timer.get_dump() for timer in self._timers],
        Task.SUBTASKS_JSON_ID: [subtask.get_id() for subtask in self.get_subtasks()]}

    def get_id(self):
        return self._id

    def get_name(self):
        return self._name

    def add_removal_callback(self, callback):
        self._removal_callbacks.append(callback)

    def remove_removal_callback(self, callback):
        if callback in self._removal_callbacks:
            self._removal_callbacks.remove(callback)

    def subtask_removed(self, subtask: Task):
        if subtask in self._subtasks:
            self._subtasks.remove(subtask)
            self._redump()

    async def timer_expired_cancelled(self, timer: Timer):
        self._timers.remove(timer)
        self._redump()
    
    async def default_timer_callback(self, timer: Timer):
        await send_message("task {} elapsed".format(self.get_name()))

    def _remove_backup(self):
        if not os.path.exists(Task._backup_path):
            return

        with Task._tasks_backup_file_lock:
            tasks = []
            with open(Task._backup_path, "r") as f:
                for line in f:
                    task = json.loads(line)
                    if task.get(Task.ID_JSON_ID) != self._id:
                        tasks.append(task)

            with open(Task._backup_path, "w") as f:
                for task in tasks:
                    f.write(Task.get_dump_str(task))


    def remove(self):
        subtasks = self.get_subtasks()
        while subtasks:
            subtask = subtasks[0]
            subtask.remove()

        self._remove_backup()

        with Task.tasks_lock:
            Task.tasks.remove(self)
        
        for callback in self._removal_callbacks:
            callback(self)

    def _redump(self):
        if self._backup:
            self._remove_backup()
            self._checked_dump()

    def get_succ_timer_callbacks(self):
        return self._shared_timer_callbacks + self._successful_timer_callbacks

    def get_cancel_timer_callbacks(self):
        return self._shared_timer_callbacks + self._cancel_timer_callbacks

    @staticmethod
    def get_dump_str(obj: dict):
        return json.dumps(obj) + "\n"

    @staticmethod
    async def check_and_create_task(task_obj: dict, backup=True):
        task = await Task._lazy_load_task(task_obj.get(Task.ID_JSON_ID)) ## needed because subtasks are loosely linked in backup (via id)
        task._name = task_obj.get(Task.NAME_JSON_ID)
        task._state = task_obj.get(Task.STATE_JSON_ID)
        task._backup = backup

        timers = task_obj.get(Task.TIMERS_JSON_ID)
        for timer in timers:
            created_timer = await Timer.check_and_create_timer(timer, backup=False, successfull_callbacks=task.get_succ_timer_callbacks(), cancel_callbacks=task.get_cancel_timer_callbacks())
            if created_timer is not None:
                created_timer.start()
                task.link_timer(created_timer)

        subtask_ids = task_obj.get(Task.SUBTASKS_JSON_ID)
        for subtask_id in subtask_ids:
            task.link_subtask(await Task._lazy_load_task(subtask_id))

        task._redump()

        return task

    @staticmethod
    async def _lazy_load_task(task_id: str):
        task = Task.get_from_id(task_id)
        if task is None:
            task = Task(None, task_id, False)
        
        return task

    @staticmethod
    async def restartExisting():
        if not os.path.exists(Task._backup_path):
            return

        tasks = []
        with Task._tasks_backup_file_lock:
            with open(Task._backup_path, "r") as f:
                for line in f:
                    tasks.append(json.loads(line))

            with open(Task._backup_path, "r+") as f:
                f.truncate(0)

        for task in tasks:
            await Task.check_and_create_task(task)

    @staticmethod
    def get_from_id(task_id: str):
        with Task.tasks_lock:
            for task in Task.tasks:
                if task.get_id() == task_id:
                    return task

        return None

    @staticmethod
    def dump_all():
        with Task.tasks_lock:
            for task in Task.tasks:
                task._redump()

    @staticmethod
    def task_state_to_html(state: str):
        return "[x]" if state == Task.STATE_TICKED else "[ ]"
