import TaskGroup
import threading
import os
from Task import Task
import uuid
import json

class TaskGroup():
    ID_JSON_ID = "id"
    NAME_JSON_ID = "name"
    SUBTASKGROUPS_JSON_ID = "subtaskgroups"
    TASKS_JSON_ID = "tasks"

    tasksGroups_lock: threading.Lock = threading.Lock()
    taskGroups: list[TaskGroup] = []

    _tasksGroups_backup_file_lock: threading.Lock = threading.Lock()
    _backup_path: str = os.environ["TASKGROUPS_PATH"]

    _subTaskGroups: list[TaskGroup] = None
    _tasks: list[Task] = None
    _removal_callbacks: list = None
    _id: str = None
    _name: str = None
    _backup: bool = None

    def __init__(self, name: str, id: str=None, backup: bool=True) -> None:
        self._subTaskGroups = []
        self._tasks = []
        self._removal_callbacks = []

        self._name = name
        self._id = id if id is not None else str(uuid.uuid4())
        self._backup = backup

        self._redump()
        TaskGroup._add_taskGroup(self)

    def _redump(self):
        if self._backup:
            self._remove_backup()
            self._checked_dump()

    def _remove_backup(self):
        if not os.path.exists(TaskGroup._backup_path):
            return

        with TaskGroup._tasksGroups_backup_file_lock:
            taskGroups = []
            with open(TaskGroup._backup_path, "r") as f:
                for line in f:
                    taskGroup = json.loads(line)
                    if taskGroup.get(Task.ID_JSON_ID) != self._id:
                        taskGroups.append(taskGroup)

            with open(TaskGroup._backup_path, "w") as f:
                for taskGroup in taskGroups:
                    f.write(TaskGroup.get_dump_str(taskGroup))

    def _checked_dump(self):
        if self._backup:
            self._dump()

    def _dump(self):
        dump_str = TaskGroup.get_dump_str(self.get_dump())
        with TaskGroup._tasksGroups_backup_file_lock:
            with open(TaskGroup._backup_path, "a+") as f:
                f.write(dump_str)

    def get_dump(self):
        return {TaskGroup.ID_JSON_ID: self.get_id(),
        TaskGroup.NAME_JSON_ID: self.get_name(),
        TaskGroup.SUBTASKGROUPS_JSON_ID: [taskGroup.get_id() for taskGroup in self.get_subTaskGroups()],
        TaskGroup.TASKS_JSON_ID: [task.get_id() for task in self.get_tasks()]}

    def get_id(self):
        return self._id

    def get_name(self):
        return self._name

    def get_subTaskGroups(self):
        return self._subTaskGroups

    def get_tasks(self):
        return self._tasks

    def add_subTaskGroup(self, name: str, id: str=None):
        subTaskGroup = TaskGroup(name, id, self._backup)
        subTaskGroup.add_removal_callback(self.subTaskGroup_removed)
        subTaskGroup._redump()

        self._subTaskGroups.append(subTaskGroup)
        self._redump()

        return subTaskGroup

    def link_subTaskGroup(self, subTaskGroup: TaskGroup):
        if subTaskGroup in self.get_subTaskGroups():
            return

        subTaskGroup._backup = self._backup
        subTaskGroup.add_removal_callback(self.subTaskGroup_removed)
        subTaskGroup._redump()

        self._subTaskGroups.append(subTaskGroup)
        self._redump()

    def unlink_subTaskGroup(self, subTaskGroup: TaskGroup):
        if subTaskGroup in self._subTaskGroups:
            subTaskGroup.remove_removal_callback(self.subTaskGroup_removed)
            self._subTaskGroups.remove(subTaskGroup)
            self._redump()

    def add_task(self, name: str, id: str=None):
        task = Task(name, id, self._backup)
        task.add_removal_callback(self.task_removed)

        self._tasks.append(task)
        self._redump()
        return task

    def link_task(self, task: Task):
        if task in self.get_tasks():
            return

        task._backup = self._backup
        task.add_removal_callback(self.task_removed)
        task._redump()

        self._tasks.append(task)
        self._redump()

    def unlink_task(self, task: Task):
        if task in self.get_tasks():
            task.remove_removal_callback(self.task_removed)
            self._tasks.remove(task)
            self._redump()

    def add_removal_callback(self, callback):
        self._removal_callbacks.append(callback)

    def remove_removal_callback(self, callback):
        if callback in self._removal_callbacks:
            self._removal_callbacks.remove(callback)

    def subTaskGroup_removed(self, subTaskGroup: TaskGroup):
        if subTaskGroup in self._subTaskGroups:
            self._subTaskGroups.remove(subTaskGroup)
            self._redump()

    def task_removed(self, task: Task):
        if task in self._tasks:
            self._tasks.remove(task)
            self._redump()

    def remove(self):
        subTaskGroups = self.get_subTaskGroups()
        while subTaskGroups:
            subTaskGroups[0].remove()

        tasks = self.get_tasks()
        while tasks:
            tasks[0].remove()

        self._remove_backup()
        TaskGroup._remove_taskGroup(self)

        for callback in self._removal_callbacks:
            callback(self)


    @staticmethod
    def get_dump_str(obj: dict):
        return json.dumps(obj) + "\n"        

    @staticmethod
    def _remove_taskGroup(taskGroup: TaskGroup):
        with TaskGroup.tasksGroups_lock:
            if taskGroup in TaskGroup.taskGroups:
                TaskGroup.taskGroups.remove(taskGroup)

    @staticmethod 
    def _add_taskGroup(taskGroup: TaskGroup):
        with TaskGroup.tasksGroups_lock:
            if not taskGroup in TaskGroup.taskGroups:
                TaskGroup.taskGroups.append(taskGroup)

    @staticmethod
    async def check_and_create_taskGroup(taskGroup_obj: dict, backup=True):
        taskGroup = await TaskGroup._lazy_load_taskGroup(taskGroup_obj.get(TaskGroup.ID_JSON_ID)) ## needed because subTasksGroups are loosely linked in backup (via id) and therefore the subtask object may needed before it's actual creation
        taskGroup._name = taskGroup_obj.get(TaskGroup.NAME_JSON_ID)
        taskGroup._backup = backup

        subTaskGroup_ids = taskGroup_obj.get(TaskGroup.SUBTASKGROUPS_JSON_ID)
        for subTaskGroup_id in subTaskGroup_ids:
            taskGroup.link_subTaskGroup(await TaskGroup._lazy_load_taskGroup(subTaskGroup_id))

        task_ids = taskGroup_obj.get(TaskGroup.TASKS_JSON_ID)
        for task_id in task_ids:
            taskGroup.link_task(await Task._lazy_load_task(task_id)) 

        taskGroup._redump()       

        return taskGroup
    
    @staticmethod
    async def _lazy_load_taskGroup(taskGroup_id: str):
        taskGroup = TaskGroup.get_from_id(taskGroup_id)
        if taskGroup is None:
            taskGroup = TaskGroup(None, taskGroup_id, backup=False)
        
        return taskGroup

    # here
    @staticmethod
    async def restartExisting():
        if not os.path.exists(TaskGroup._backup_path):
            return

        taskGroups = []
        with TaskGroup._tasksGroups_backup_file_lock:
            with open(TaskGroup._backup_path, "r") as f:
                for line in f:
                    taskGroups.append(json.loads(line))

            with open(TaskGroup._backup_path, "r+") as f:
                f.truncate(0)

        for taskGroup in taskGroups:
            await TaskGroup.check_and_create_taskGroup(taskGroup)

    @staticmethod
    def get_from_id(taskGroup_id: str):
        with TaskGroup.tasksGroups_lock:
            for taskGroup in TaskGroup.taskGroups:
                if taskGroup.get_id() == taskGroup_id:
                    return taskGroup

        return None

    @staticmethod
    def dump_all():
        with TaskGroup.tasksGroups_lock:
            for taskGroup in TaskGroup.taskGroups:
                taskGroup._redump()